# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Hale Ligh <michael.ligh@mnin.org>
#
# Contributors/References:
#   Richard Stevens and Eoghan Casey
#   Extracting Windows Cmd Line Details from Physical Memory.
#   http://ww.dfrws.org/2010/proceedings/stevens.pdf
#   Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

from rekall import obj
from rekall import scan
from rekall.plugins.overlays import basic
from rekall.plugins.windows import common
from rekall.plugins.windows import vadinfo

MAX_HISTORY_DEFAULT = 50
HISTORY_BUFFERS_DEFAULTS = 4

#--------------------------------------------------------------------------------
# VTypes
#--------------------------------------------------------------------------------

# The following all use these standard types.
common_types = {
    '_LIST_ENTRY' : [ 0x8, {
            'Flink' : [ 0x0, ['pointer', ['_LIST_ENTRY']]],
            'Blink' : [ 0x4, ['pointer', ['_LIST_ENTRY']]],
            }]}

common_types_64 = {
    '_LIST_ENTRY' : [ 0x10, {
            'Flink' : [ 0x0, ['pointer', ['_LIST_ENTRY']]],
            'Blink' : [ 0x8, ['pointer', ['_LIST_ENTRY']]],
            }]}

# Windows 7 Types from conhost.exe
conhost_types_x86 = {
    '_COMMAND': [ None, {
    'CmdLength': [ 0x00, ['unsigned short']],
    'Cmd' : [ 0x02, ['UnicodeString', dict(encoding = 'utf16',
                                           length = lambda x : x.CmdLength)]],
    }],
    '_COMMAND_HISTORY': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'Flags' : [ 0x08, ['Flags', {'bitmap': {'Allocated': 0, 'Reset': 1}}]],
    'Application': [ 0x0C, ['pointer', ['UnicodeString', dict(encoding = 'utf16',
                                                              length = 256)]]],
    'CommandCount': [ 0x10, ['short']],
    'LastAdded': [ 0x12, ['short']],
    'LastDisplayed': [ 0x14, ['short']],
    'FirstCommand': [ 0x16, ['short']],
    'CommandCountMax': [ 0x18, ['short']],
    'ProcessHandle': [ 0x1C, ['unsigned int']],
    'PopupList': [ 0x20, ['_LIST_ENTRY']],
    'CommandBucket': [ 0x28, ['array', lambda x : x.CommandCount, ['pointer', ['_COMMAND']]]],
    }],
    '_ALIAS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'SourceLength': [ 0x08, ['unsigned short']],
    'TargetLength': [ 0x0A, ['unsigned short']],
    'Source': [ 0x0C, ['pointer', ['UnicodeString', dict(encoding = 'utf16',
                                                         length = lambda x : x.SourceLength * 2)]]],
    'Target': [ 0x10, ['pointer', ['UnicodeString', dict(encoding = 'utf16',
                                                         length = lambda x : x.TargetLength * 2)]]],
    }],
    '_EXE_ALIAS_LIST' : [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ExeLength': [ 0x08, ['unsigned short']],
    'ExeName': [ 0x0C, ['pointer', ['UnicodeString', dict(encoding = 'utf16',
                                                          length = lambda x : x.ExeLength * 2)]]],
    'AliasList': [ 0x10, ['_LIST_ENTRY']],
    }],
    '_POPUP_LIST' : [ None, {
    'ListEntry' : [ 0x00, ['_LIST_ENTRY']],
    }],
    '_CONSOLE_INFORMATION': [ None, {
    'CurrentScreenBuffer': [ 0x98, ['pointer', ['_SCREEN_INFORMATION']]],
    'ScreenBuffer': [ 0x9C, ['pointer', ['_SCREEN_INFORMATION']]],
    'HistoryList': [ 0xD4, ['_LIST_ENTRY']],
    'ProcessList': [ 0x18, ['_LIST_ENTRY']], # GetConsoleProcessList()
    'ExeAliasList': [ 0xDC, ['_LIST_ENTRY']], # GetConsoleAliasExes()
    'HistoryBufferCount': [ 0xE4, ['unsigned short']], # GetConsoleHistoryInfo()
    'HistoryBufferMax': [ 0xE6, ['unsigned short']], # GetConsoleHistoryInfo()
    'CommandHistorySize': [ 0xE8, ['unsigned short']],
    'OriginalTitle': [ 0xEC, ['pointer', ['UnicodeString', dict(encoding = 'utf16',
                                                                length = 256)]]], # GetConsoleOriginalTitle()
    'Title': [ 0xF0, ['pointer', ['UnicodeString', dict(encoding = 'utf16',
                                                        length = 256)]]], # GetConsoleTitle()
    }],
    '_CONSOLE_PROCESS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ProcessHandle': [ 0x8, ['unsigned int']],
    }],
    '_SCREEN_INFORMATION': [ None, {
    'ScreenX': [ 0x08, ['short']],
    'ScreenY': [ 0x0A, ['short']],
    'Rows': [ 0x3C, ['pointer', ['array', lambda x : x.ScreenY, ['_ROW']]]],
    'Next': [ 0xDC, ['pointer', ['_SCREEN_INFORMATION']]],
    }],
    '_ROW': [ 0x1C, {
    'Chars': [ 0x08, ['pointer', ['UnicodeString', dict(encoding = 'utf16',
                                                        length = 256)]]],
    }],
}

# Windows 7 Types from conhost.exe
conhost_types_x64 = {
    '_COMMAND': [ None, {
    'CmdLength': [ 0x00, ['unsigned short']],
    'Cmd' : [ 0x02, ['UnicodeString', dict(
                        encoding='utf16',
                        length=lambda x : x.CmdLength)]],
    }],
    '_COMMAND_HISTORY': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    # AllocateCommandHistory()
    'Flags' : [ 0x10, ['Flags', {'bitmap': {'Allocated': 0, 'Reset': 1}}]],

    # AllocateCommandHistory()
    'Application': [ 0x18, ['pointer', ['UnicodeString', dict(
                            encoding='utf16', length = 256)]]],
    'CommandCount': [ 0x20, ['short']],
    'LastAdded': [ 0x22, ['short']],
    'LastDisplayed': [ 0x24, ['short']],
    'FirstCommand': [ 0x26, ['short']],
    'CommandCountMax': [ 0x28, ['short']], # AllocateCommandHistory()
    'ProcessHandle': [ 0x30, ['address']], # AllocateCommandHistory()
    'PopupList': [ 0x38, ['_LIST_ENTRY']], # AllocateCommandHistory()
    'CommandBucket': [ 0x48, ['Array', dict(
                        count=lambda x : x.CommandCount,
                        target='Pointer',
                        target_args=dict(
                            target='_COMMAND')
                        )]],
    }],
    '_ALIAS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'SourceLength': [ 0x10, ['unsigned short']], # AddAlias()
    'TargetLength': [ 0x12, ['unsigned short']], # AddAlias()

    # reversed from AddAlias()
    'Source': [ 0x18, ['pointer', ['UnicodeString', dict(
                            encoding='utf16',
                            length=lambda x : x.SourceLength * 2)]]],

    'Target': [ 0x20, ['pointer', ['UnicodeString', dict(
                            encoding='utf16',
                            length = lambda x : x.TargetLength * 2)]]], # AddAlias()
    }],
    '_EXE_ALIAS_LIST' : [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ExeLength': [ 0x10, ['unsigned short']], # AddExeAliasList()

    # AddExeAliasList()
    'ExeName': [ 0x18, ['pointer', ['UnicodeString', dict(
                            encoding='utf16',
                            length=lambda x : x.ExeLength * 2)]]],

    'AliasList': [ 0x20, ['_LIST_ENTRY']], # AddExeAliasList()
    }],
    '_POPUP_LIST' : [ None, {
    'ListEntry' : [ 0x00, ['_LIST_ENTRY']],
    }],
    '_CONSOLE_INFORMATION': [ None, {
    'ProcessList': [ 0x28, ['_LIST_ENTRY']], # SrvGetConsoleProcessList()
    'CurrentScreenBuffer': [ 0xE0, ['pointer', ['_SCREEN_INFORMATION']]], # AllocateConsole()
    'ScreenBuffer': [ 0xE8, ['pointer', ['_SCREEN_INFORMATION']]], # AllocateConsole()
    'HistoryList': [ 0x148, ['_LIST_ENTRY']], # AllocateCommandHistory()
    'ExeAliasList': [ 0x148, ['_LIST_ENTRY']], # SrvGetConsoleAliasExes()
    'HistoryBufferCount': [ 0x168, ['unsigned short']], # AllocateConsole()
    'HistoryBufferMax': [ 0x16A, ['unsigned short']], # AllocateConsole()
    'CommandHistorySize': [ 0x16C, ['unsigned short']], # AllocateConsole()
    'OriginalTitle': [ 0x170, ['pointer', ['UnicodeString', dict(
                            encoding='utf16', length=256)]]], # SrvGetConsoleTitle()
    'Title': [ 0x178, ['pointer', ['UnicodeString', dict(
                            encoding='utf16', length=256)]]], # SrvGetConsoleTitle()
    }],
    '_CONSOLE_PROCESS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ProcessHandle': [ 0x10, ['unsigned int']], # FindProcessInList()
    }],
    '_SCREEN_INFORMATION': [ None, {
    'ScreenX': [ 8, ['short']],
    'ScreenY': [ 10, ['short']],
    'Rows': [ 0x48, ['pointer', ['array', lambda x : x.ScreenY, ['_ROW']]]],
    'Next': [ 0x128, ['pointer', ['_SCREEN_INFORMATION']]],
    }],
    '_ROW': [ 0x28, {
    'Chars': [ 0x08, ['pointer', ['UnicodeString', dict(
                            encoding='utf16', length=256)]]],
    }],
}

# Windows XP, 2003, 2008, Vista from winsrv.dll
winsrv_types_x86 = {
    '_COMMAND': [ None, {
    'CmdLength': [ 0x00, ['unsigned short']],
    'Cmd' : [ 0x02, ['UnicodeString', dict(
                        encoding='utf16',
                        length=lambda x : x.CmdLength)]],
    }],
    '_COMMAND_HISTORY': [ None, {
    'Flags' : [ 0x00, ['Flags', {'bitmap': {'Allocated': 0, 'Reset': 1}}]],
    'ListEntry': [ 0x04, ['_LIST_ENTRY']],
    'Application': [ 0x0C, ['pointer', ['UnicodeString', dict(
                            encoding='utf16', length=256)]]],
    'CommandCount': [ 0x10, ['short']],
    'LastAdded': [ 0x12, ['short']],
    'LastDisplayed': [ 0x14, ['short']],
    'FirstCommand': [ 0x16, ['short']],
    'CommandCountMax': [ 0x18, ['short']],
    'ProcessHandle': [ 0x1C, ['unsigned int']],
    'PopupList': [ 0x20, ['_LIST_ENTRY']],
    'CommandBucket': [ 0x28, ['array', lambda x : x.CommandCount, ['pointer', ['_COMMAND']]]],
    }],
    '_ALIAS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'SourceLength': [ 0x08, ['unsigned short']],
    'TargetLength': [ 0x0A, ['unsigned short']],
    'Source': [ 0x0C, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = lambda x : x.SourceLength * 2)]]],
    'Target': [ 0x10, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = lambda x : x.TargetLength * 2)]]],
    }],
    '_EXE_ALIAS_LIST' : [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ExeLength': [ 0x08, ['unsigned short']],
    'ExeName': [ 0x0C, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = lambda x : x.ExeLength * 2)]]],
    'AliasList': [ 0x10, ['_LIST_ENTRY']],
    }],
    '_POPUP_LIST' : [ None, {
    'ListEntry' : [ 0x00, ['_LIST_ENTRY']],
    }],
    '_CONSOLE_INFORMATION': [ None, {
    'CurrentScreenBuffer': [ 0xB0, ['pointer', ['_SCREEN_INFORMATION']]],
    'ScreenBuffer': [ 0xB4, ['pointer', ['_SCREEN_INFORMATION']]],
    'HistoryList': [ 0x108, ['_LIST_ENTRY']],
    'ProcessList': [ 0x100, ['_LIST_ENTRY']],
    'ExeAliasList': [ 0x110, ['_LIST_ENTRY']],
    'HistoryBufferCount': [ 0x118, ['unsigned short']],
    'HistoryBufferMax': [ 0x11A, ['unsigned short']],
    'CommandHistorySize': [ 0x11C, ['unsigned short']],
    'OriginalTitle': [ 0x124, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = 256)]]],
    'Title': [ 0x128, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = 256)]]],
    }],
    '_CONSOLE_PROCESS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ProcessHandle': [ 0x08, ['unsigned int']],
    'Process': [ 0x0C, ['pointer', ['_CSR_PROCESS']]],
    }],
    '_SCREEN_INFORMATION': [ None, {
    'Console': [ 0x00, ['pointer', ['_CONSOLE_INFORMATION']]],
    'ScreenX': [ 0x24, ['short']],
    'ScreenY': [ 0x26, ['short']],
    'Rows': [ 0x58, ['pointer', ['array', lambda x : x.ScreenY, ['_ROW']]]],
    'Next': [ 0xF8, ['pointer', ['_SCREEN_INFORMATION']]],
    }],
    '_ROW': [ 0x1C, {
    'Chars': [ 0x08, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = 256)]]],
    }],
    '_CSR_PROCESS' : [ 0x60, { # this is a public PDB
    'ClientId' : [ 0x0, ['_CLIENT_ID']],
    'ListLink' : [ 0x8, ['_LIST_ENTRY']],
    'ThreadList' : [ 0x10, ['_LIST_ENTRY']],
    'NtSession' : [ 0x18, ['pointer', ['_CSR_NT_SESSION']]],
    'ClientPort' : [ 0x1c, ['pointer', ['void']]],
    'ClientViewBase' : [ 0x20, ['pointer', ['unsigned char']]],
    'ClientViewBounds' : [ 0x24, ['pointer', ['unsigned char']]],
    'ProcessHandle' : [ 0x28, ['pointer', ['void']]],
    'SequenceNumber' : [ 0x2c, ['unsigned long']],
    'Flags' : [ 0x30, ['unsigned long']],
    'DebugFlags' : [ 0x34, ['unsigned long']],
    'ReferenceCount' : [ 0x38, ['unsigned long']],
    'ProcessGroupId' : [ 0x3c, ['unsigned long']],
    'ProcessGroupSequence' : [ 0x40, ['unsigned long']],
    'LastMessageSequence' : [ 0x44, ['unsigned long']],
    'NumOutstandingMessages' : [ 0x48, ['unsigned long']],
    'ShutdownLevel' : [ 0x4c, ['unsigned long']],
    'ShutdownFlags' : [ 0x50, ['unsigned long']],
    'Luid' : [ 0x54, ['_LUID']],
    'ServerDllPerProcessData' : [ 0x5c, ['array', 1, ['pointer', ['void']]]],
    }],
}

winsrv_types_x64 = {
    '_COMMAND': [ None, {
    'CmdLength': [ 0x00, ['unsigned short']],
    'Cmd' : [ 0x02, ['UnicodeString', dict(encoding = 'utf16', length = lambda x : x.CmdLength)]],
    }],
    '_COMMAND_HISTORY': [ None, {
    'Flags' : [ 0x00, ['Flags', {'bitmap': {'Allocated': 0, 'Reset': 1}}]],
    'ListEntry': [ 0x08, ['_LIST_ENTRY']],
    'Application': [ 0x18, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = 256)]]],
    'CommandCount': [ 0x20, ['short']],
    'LastAdded': [ 0x22, ['short']],
    'LastDisplayed': [ 0x24, ['short']],
    'FirstCommand': [ 0x26, ['short']],
    'CommandCountMax': [ 0x28, ['short']],
    'ProcessHandle': [ 0x30, ['unsigned int']],
    'PopupList': [ 0x38, ['_LIST_ENTRY']],
    'CommandBucket': [ 0x48, ['array', lambda x : x.CommandCount, ['pointer', ['_COMMAND']]]],
    }],
    '_ALIAS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'SourceLength': [ 0x10, ['unsigned short']],
    'TargetLength': [ 0x12, ['unsigned short']],
    'Source': [ 0x14, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = lambda x : x.SourceLength * 2)]]],
    'Target': [ 0x1C, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = lambda x : x.TargetLength * 2)]]],
    }],
    '_EXE_ALIAS_LIST' : [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ExeLength': [ 0x10, ['unsigned short']],
    'ExeName': [ 0x12, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = lambda x : x.ExeLength * 2)]]],
    'AliasList': [ 0x1A, ['_LIST_ENTRY']],
    }],
    '_POPUP_LIST' : [ None, {
    'ListEntry' : [ 0x00, ['_LIST_ENTRY']],
    }],
    '_CONSOLE_INFORMATION': [ None, {
    'CurrentScreenBuffer': [ 0xE8, ['pointer', ['_SCREEN_INFORMATION']]],
    'ScreenBuffer': [ 0xF0, ['pointer', ['_SCREEN_INFORMATION']]],
    'HistoryList': [ 0x188, ['_LIST_ENTRY']],
    'ProcessList': [ 0x178, ['_LIST_ENTRY']],
    'ExeAliasList': [ 0x198, ['_LIST_ENTRY']],
    'HistoryBufferCount': [ 0x1A8, ['unsigned short']],
    'HistoryBufferMax': [ 0x1AA, ['unsigned short']],
    'CommandHistorySize': [ 0x1AC, ['unsigned short']],
    'OriginalTitle': [ 0x1B0, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = 256)]]],
    'Title': [ 0x1B8, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = 256)]]],
    }],
    '_CONSOLE_PROCESS': [ None, {
    'ListEntry': [ 0x00, ['_LIST_ENTRY']],
    'ProcessHandle': [ 0x10, ['unsigned int']],
    'Process': [ 0x18, ['pointer', ['_CSR_PROCESS']]],
    }],
    '_SCREEN_INFORMATION': [ None, {
    'Console': [ 0x00, ['pointer', ['_CONSOLE_INFORMATION']]],
    'ScreenX': [ 0x28, ['short']],
    'ScreenY': [ 0x2A, ['short']],
    'Rows': [ 0x68, ['pointer', ['array', lambda x : x.ScreenY, ['_ROW']]]],
    'Next': [ 0x128, ['pointer', ['_SCREEN_INFORMATION']]],
    }],
    '_ROW': [ 0x28, {
    'Chars': [ 0x08, ['pointer', ['UnicodeString', dict(encoding = 'utf16', length = 256)]]],
    }],
    '_CSR_PROCESS' : [ 0x60, { # this is a public PDB
    'ClientId' : [ 0x0, ['_CLIENT_ID']],
    'ListLink' : [ 0x8, ['_LIST_ENTRY']],
    'ThreadList' : [ 0x10, ['_LIST_ENTRY']],
    'NtSession' : [ 0x18, ['pointer', ['_CSR_NT_SESSION']]],
    'ClientPort' : [ 0x1c, ['pointer', ['void']]],
    'ClientViewBase' : [ 0x20, ['pointer', ['unsigned char']]],
    'ClientViewBounds' : [ 0x24, ['pointer', ['unsigned char']]],
    'ProcessHandle' : [ 0x28, ['pointer', ['void']]],
    'SequenceNumber' : [ 0x2c, ['unsigned long']],
    'Flags' : [ 0x30, ['unsigned long']],
    'DebugFlags' : [ 0x34, ['unsigned long']],
    'ReferenceCount' : [ 0x38, ['unsigned long']],
    'ProcessGroupId' : [ 0x3c, ['unsigned long']],
    'ProcessGroupSequence' : [ 0x40, ['unsigned long']],
    'LastMessageSequence' : [ 0x44, ['unsigned long']],
    'NumOutstandingMessages' : [ 0x48, ['unsigned long']],
    'ShutdownLevel' : [ 0x4c, ['unsigned long']],
    'ShutdownFlags' : [ 0x50, ['unsigned long']],
    'Luid' : [ 0x54, ['_LUID']],
    'ServerDllPerProcessData' : [ 0x5c, ['array', 1, ['pointer', ['void']]]],
    }],
}

#--------------------------------------------------------------------------------
# Object Classes
#--------------------------------------------------------------------------------

class _CONSOLE_INFORMATION(obj.Struct):
    """ object class for console information structs """

    def get_screens(self):
        """Generator for screens in the console.

        A console can have multiple screen buffers at a time,
        but only the current/active one is displayed.

        Multiple screens are tracked using the singly-linked
        list _SCREEN_INFORMATION.Next.

        See CreateConsoleScreenBuffer
        """
        screens = [self.CurrentScreenBuffer]

        if self.ScreenBuffer not in screens:
            screens.append(self.ScreenBuffer)

        for screen in screens:
            cur = screen
            while cur and cur.v() != 0:
                yield cur
                cur = cur.Next.dereference()

class _SCREEN_INFORMATION(obj.Struct):
    """ object class for screen information """

    def get_buffer(self, truncate = True):
        """Get the screen buffer.

        The screen buffer is comprised of the screen's Y
        coordinate which tells us the number of rows and
        the X coordinate which tells us the width of each
        row in characters. These together provide all of
        the input and output that users see when the
        console is displayed.

        @param truncate: True if the empty rows at the
        end (i.e. bottom) of the screen buffer should be
        supressed.
        """
        rows = []

        for _, row in enumerate(self.Rows.dereference()):
            if row.Chars.is_valid():
                rows.append(str(row.Chars.dereference())[0:self.ScreenX])

        # To truncate empty rows at the end, walk the list
        # backwards and get the last non-empty row. Use that
        # row index to splice. An "empty" row isn't just ""
        # as one might assume. It is actually ScreenX number
        # of space characters

        if truncate:
            non_empty_index = 0
            for index, row in enumerate(reversed(rows)):
                ## It seems that when the buffer width is greater than 128
                ## characters, its truncated to 128 in memory.
                if row.count(" ") != min(self.ScreenX, 128):
                    non_empty_index = index
                    break
            if non_empty_index == 0:
                rows = []
            else:
                rows = rows[0:len(rows) - non_empty_index]

        return rows

#--------------------------------------------------------------------------------
# Application Profiles
#--------------------------------------------------------------------------------
class WinSrv86(basic.Profile32Bits, basic.BasicWindowsClasses):
    """A domain specific profile for the xp, 2008."""

    def __init__(self, **kwargs):
        super(WinSrv86, self).__init__(**kwargs)
        self.add_types(common_types)
        self.add_types(winsrv_types_x86)
        self.add_classes({
                '_CONSOLE_INFORMATION': _CONSOLE_INFORMATION,
                '_SCREEN_INFORMATION': _SCREEN_INFORMATION,
                })


class WinSrv64(basic.ProfileLLP64, basic.BasicWindowsClasses):
    """A domain specific profile for the xp, 2008."""

    def __init__(self, **kwargs):
        super(WinSrv64, self).__init__(**kwargs)
        self.add_types(common_types_64)
        self.add_types(winsrv_types_x64)
        self.add_classes({
                '_CONSOLE_INFORMATION': _CONSOLE_INFORMATION,
                '_SCREEN_INFORMATION': _SCREEN_INFORMATION,
                })


class ConHost86(basic.Profile32Bits, basic.BasicWindowsClasses):
    """A domain specific profile for windows 7."""

    def __init__(self, **kwargs):
        super(ConHost86, self).__init__(**kwargs)
        self.add_types(common_types)
        self.add_types(conhost_types_x86)
        self.add_classes({
                '_CONSOLE_INFORMATION': _CONSOLE_INFORMATION,
                '_SCREEN_INFORMATION': _SCREEN_INFORMATION,
                })


class ConHost64(basic.ProfileLLP64, basic.BasicWindowsClasses):
    """A domain specific profile for windows 7."""

    def __init__(self, **kwargs):
        super(ConHost64, self).__init__(**kwargs)
        self.add_types(common_types_64)
        self.add_types(conhost_types_x64)
        self.add_classes({
                '_CONSOLE_INFORMATION': _CONSOLE_INFORMATION,
                '_SCREEN_INFORMATION': _SCREEN_INFORMATION,
                })


#--------------------------------------------------------------------------------
# CmdScan Plugin
#--------------------------------------------------------------------------------
class HistoryScanner(vadinfo.VadScanner):
    """A vad scanner for command histories.

    The default pattern we search for, as described by Stevens and Casey,
    is "\x32\x00". That's because CommandCountMax is a little-endian
    unsigned short whose default value is 50. However, that value can be
    changed by right clicking cmd.exe and going to Properties->Options->Cmd History
    or by calling the API function kernel32!SetConsoleHistoryInfo. Thus
    you can tweak the search criteria by using the --MAX_HISTORY.
    """
    def __init__(self, max_history=MAX_HISTORY_DEFAULT, **kwargs):
        super(HistoryScanner, self).__init__(**kwargs)
        self.max_history = max_history

    def scan(self, **kwargs):
        for hit in super(HistoryScanner, self).scan(**kwargs):
            # Check to see if the object is valid.
            hist = self.profile.Object(
                "_COMMAND_HISTORY", vm = self.address_space,
                offset = hit - self.profile.get_obj_offset("_COMMAND_HISTORY",
                                                           "CommandCountMax"))

            if not hist.is_valid(): continue

            # The count must be between zero and max
            if hist.CommandCount < 0 or hist.CommandCount > self.max_history:
                continue

            # Last added must be between -1 and max
            if hist.LastAdded < -1 or hist.LastAdded > self.max_history:
                continue

            # Last displayed must be between -1 and max
            if hist.LastDisplayed < -1 or hist.LastDisplayed > self.max_history:
                continue

            # First command must be between zero and max
            if hist.FirstCommand < 0 or hist.FirstCommand > self.max_history:
                continue

            # Validate first command with last added
            if hist.FirstCommand != 0 and hist.FirstCommand != hist.LastAdded + 1:
                continue

            # Process handle must be a valid pid
            if hist.ProcessHandle <= 0 or hist.ProcessHandle > 0xFFFF:
                continue

            Popup = self.profile._POPUP_LIST(offset = hist.PopupList.Flink,
                                             vm = self.address_space)

            # Check that the popup list entry is in tact
            if Popup.ListEntry.Blink != hist.PopupList.obj_offset:
                continue

            yield hist


class CmdScan(common.WindowsCommandPlugin):
    """Extract command history by scanning for _COMMAND_HISTORY"""
    __name = "cmdscan"

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(CmdScan, cls).args(parser)
        parser.add_argument("--max_history", default=MAX_HISTORY_DEFAULT,
                            help="Value of history buffer size. See "
                            "HKEY_CURRENT_USER\\Console\\HistoryBufferSize "
                            "for default.")

    def __init__(self, max_history=MAX_HISTORY_DEFAULT, **kwargs):
        """Scan for _COMMAND_HISTORY structs.

        Args:
           max_history: The maximum size of the command history buffer. This
              parameter is configurable but is required for scanning.
        """
        super(CmdScan, self).__init__(**kwargs)
        self.max_history = max_history

    def generate_hits(self):
        """Generates _COMMAND_HISTORY objects."""
        memory_model = self.profile.metadata("memory_model")
        # The process we select is conhost on Win7 or csrss for others

        if self.profile.metadata("major") >= 6:
            process_name = "conhost.exe"
            process_profile = ConHost64() if memory_model=="64bit" else ConHost86()
        else:
            process_name = "csrss.exe"
            process_profile = WinSrv64() if memory_model=="64bit" else WinSrv86()

        # Only select those processes we care about:
        for task in self.session.plugins.pslist(
            proc_regex=process_name).filter_processes():
            scanner = HistoryScanner(task=task, process_profile=process_profile,
                                     max_history=self.max_history)

            pattern = chr(self.max_history) + "\x00"
            scanner.checks = [("MultiStringFinderCheck", dict(needles=[pattern]))]
            scanner.build_constraints()

            for hist in scanner.scan():
                yield task, hist

    def render(self, renderer):
        for task, hist in self.generate_hits():
            renderer.section()
            renderer.format(u"CommandProcess: {0} Pid: {1}\n",
                            task.ImageFileName, task.UniqueProcessId)
            renderer.format(
                u"CommandHistory: {0:#x} Application: {1} Flags: {2}\n",
                hist.obj_offset, hist.Application.dereference(), hist.Flags)

            renderer.format(
                u"CommandCount: {0} LastAdded: {1} LastDisplayed: {2}\n",
                hist.CommandCount, hist.LastAdded, hist.LastDisplayed)

            renderer.format(u"FirstCommand: {0} CommandCountMax: {1}\n",
                            hist.FirstCommand, hist.CommandCountMax)

            renderer.format(u"ProcessHandle: {0:#x}\n", hist.ProcessHandle)


            renderer.table_header([("Cmd", "command", ">3"),
                                   ("Address", "address", "[addrpad]"),
                                   ("Text", "text", "50")])

            # If the _COMMAND_HISTORY is in use, we would only take
            # hist.CommandCount but since we're brute forcing, try the
            # maximum and hope that some slots were not overwritten
            # or zero-ed out.
            pointers = hist.obj_profile.Array(
                target="address", count=hist.CommandCountMax,
                offset=hist.obj_offset + hist.obj_profile.get_obj_offset(
                    "_COMMAND_HISTORY", "CommandBucket"),
                vm=hist.obj_vm)

            for i, p in enumerate(pointers):
                cmd = p.dereference_as("_COMMAND")
                if cmd.obj_offset and cmd.Cmd:
                    renderer.table_row(i, cmd.obj_offset, cmd.Cmd)


#--------------------------------------------------------------------------------
# Consoles Plugin
#--------------------------------------------------------------------------------

class ConsoleScanner(vadinfo.VadScanner):
    """A scanner for _CONSOLE_INFORMATION."""
    def __init__(self, max_history=MAX_HISTORY_DEFAULT,
                 history_buffers=HISTORY_BUFFERS_DEFAULTS, **kwargs):
        super(ConsoleScanner, self).__init__(**kwargs)
        self.max_history = max_history
        self.history_buffers = history_buffers

    def scan(self, **kwargs):
        for hit in super(ConsoleScanner, self).scan(**kwargs):
            # Check to see if the object is valid.
            console = self.profile.Object(
                "_CONSOLE_INFORMATION", offset = hit -
                self.profile.get_obj_offset(
                    "_CONSOLE_INFORMATION", "CommandHistorySize"),
                vm = self.address_space, parent=self.task)

            if (console.HistoryBufferMax != self.history_buffers or
                console.HistoryBufferCount > self.history_buffers):
                continue

            # Check the first command history as the final constraint
            next_history = console.HistoryList.Flink.dereference().dereference_as(
                "_COMMAND_HISTORY", "ListEntry")
            if next_history.CommandCountMax != self.max_history:
                continue

            yield console


class Consoles(CmdScan):
    """Extract command history by scanning for _CONSOLE_INFORMATION"""

    __name = "consoles"

    @classmethod
    def args(cls, parser):
        """Declare the command line args we need."""
        super(Consoles, cls).args(parser)
        parser.add_argument("--history-buffers", default=HISTORY_BUFFERS_DEFAULTS,
                            help="Number of history buffere. See "
                            "HKCU\Console\NumberOfHistoryBuffers. "
                            "Uses %s by default." % HISTORY_BUFFERS_DEFAULTS)

    def __init__(self, history_buffers=HISTORY_BUFFERS_DEFAULTS, **kwargs):
        super(Consoles, self).__init__(**kwargs)
        self.history_buffers = history_buffers

    def generate_hits(self):
        """Generates _CONSOLE_INFORMATION objects."""
        memory_model = self.profile.metadata("memory_model")
        # The process we select is conhost on Win7 or csrss for others

        # Only select those processes we care about:
        for task in self.session.plugins.pslist(
            proc_regex="(conhost.exe|csrss.exe)").filter_processes():

            if str(task.ImageFileName).lower() == "conhost.exe":
                process_profile = ConHost64() if memory_model=="64bit" else ConHost86()
            elif str(task.ImageFileName).lower() == "csrss.exe":
                process_profile = WinSrv64() if memory_model=="64bit" else WinSrv86()
            else: continue

            scanner = ConsoleScanner(task=task, process_profile=process_profile,
                                     max_history=self.max_history,
                                     history_buffers=self.history_buffers)

            pattern = chr(self.max_history) + "\x00"
            scanner.checks = [("MultiStringFinderCheck", dict(needles=[pattern]))]
            scanner.build_constraints()

            for console in scanner.scan():
                yield task, console

    def render(self, renderer):
        for task, console in self.generate_hits():
            renderer.section()
            renderer.format(u"ConsoleProcess: {0} Pid: {1}\n",
                            task.ImageFileName, task.UniqueProcessId)
            renderer.format(u"Console: {0:#x} CommandHistorySize: {1}\n",
                            console, console.CommandHistorySize)
            renderer.format(u"HistoryBufferCount: {0} HistoryBufferMax: {1}\n",
                            console.HistoryBufferCount, console.HistoryBufferMax)
            renderer.format(u"OriginalTitle: {0}\n",
                            console.OriginalTitle.dereference())

            renderer.format(u"Title: {0}\n", console.Title.dereference())

            for console_proc in console.ProcessList.list_of_type(
                "_CONSOLE_PROCESS", "ListEntry"):
                process = task.ObReferenceObjectByHandle(
                    console_proc.ProcessHandle, type="_EPROCESS")

                if process:
                    renderer.format(
                        u"AttachedProcess: {0} Pid: {1} Handle: {2:#x}\n",
                        process.ImageFileName, process.UniqueProcessId,
                        console_proc.ProcessHandle)

            for hist in console.HistoryList.list_of_type(
                "_COMMAND_HISTORY", "ListEntry"):
                renderer.format(u"----\n")

                renderer.format(
                    u"CommandHistory: {0:#x} Application: {1} Flags: {2}\n",
                    hist, hist.Application.dereference(),
                    hist.Flags)
                renderer.format(
                    u"CommandCount: {0} LastAdded: {1} LastDisplayed: {2}\n",
                    hist.CommandCount, hist.LastAdded, hist.LastDisplayed)

                renderer.format(u"FirstCommand: {0} CommandCountMax: {1}\n",
                                hist.FirstCommand, hist.CommandCountMax)

                renderer.format(u"ProcessHandle: {0:#x}\n", hist.ProcessHandle)
                for i, cmd in enumerate(hist.CommandBucket):
                    if cmd.Cmd.is_valid():
                        renderer.format(u"Cmd #{0} at {1:#x}: {2}\n",
                                        i, cmd, str(cmd.Cmd))

            for exe_alias in console.ExeAliasList.list_of_type(
                "_EXE_ALIAS_LIST", "ListEntry"):

                for alias in exe_alias.AliasList.list_of_type(
                    "_ALIAS", "ListEntry"):
                    renderer.format(u"----\n")
                    renderer.format(
                        u"Alias: {0} Source: {1} Target: {2}\n",
                        exe_alias.ExeName.dereference(), alias.Source.dereference(),
                        alias.Target.dereference())

            for screen in console.get_screens():
                renderer.format(u"----\n")
                renderer.format(u"Screen {0:#x} X:{1} Y:{2}\n",
                                screen.dereference(), screen.ScreenX,
                                screen.ScreenY)

                renderer.format(u"Dump:\n{0}\n", '\n'.join(screen.get_buffer()))



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Tests for the yarascan plugins."""

from rekall import testlib


class TestYara(testlib.RekallBaseUnitTestCase):
    """Test the yarascan module."""

    PARAMETERS = dict(commandline="yarascan --string sys")

    def testYara(self):
        previous = self.baseline['output']
        current = self.current['output']

        self.assertTrue(
            self.ExtractColumn(current, 0, 2, seperator=" "),
            self.ExtractColumn(previous, 0, 2, seperator=" "))


    def XXtestYarra2(self):
        metadata = self.LoadPreviousRunData("yarascan")
        session = self.BuildUserSession("yarascan")

        str(session.plugins.pslist())

        # Instantiate the yara plugin.
        rules = metadata['yarascan_test']
        for task, rule in rules.items():
            if task:
                task_as = task.get_process_address_space()
            else:
                task_as = session.kernel_address_space

            yarascan = session.plugins.yarascan(yara_expression=rule,
                                                address_space=task_as)

            for rule, offset, _, match in yarascan.generate_hits(task_as):
                data = task_as.read(offset, len(match))
                self.assertEqual(match, data)



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import bisect

from rekall import obj
from rekall.plugins.windows import common


class Modules(common.WindowsCommandPlugin):
    """Print list of loaded modules."""

    __name = "modules"

    # A local cache for find_modules. Key is module base and value is the
    # _LDR_DATA_TABLE_ENTRY for the module.
    mod_lookup = None
    modlist = None

    def __init__(self, **kwargs):
        """List kernel modules by walking the PsLoadedModuleList."""
        super(Modules, self).__init__(**kwargs)

    def lsmod(self):
        """ A Generator for modules (uses _KPCR symbols) """
        if not self.mod_lookup:
            self._make_cache()

        for module in self.mod_lookup.values():
            yield module

    def addresses(self):
        """Returns a list of module addresses."""
        if not self.mod_lookup:
            self._make_cache()

        return sorted(self.mod_lookup.keys())

    def _make_cache(self):
        self.mod_lookup = {}

        ## Try to iterate over the process list in PsActiveProcessHead
        ## (its really a pointer to a _LIST_ENTRY)
        PsLoadedModuleList = self.kdbg.PsLoadedModuleList.dereference_as(
            "_LIST_ENTRY", vm=self.kernel_address_space)

        for l in PsLoadedModuleList.list_of_type("_LDR_DATA_TABLE_ENTRY",
                                                 "InLoadOrderLinks"):
            self.mod_lookup[l.DllBase.v()] = l

        self.modlist = sorted(self.mod_lookup.keys())

    def find_module(self, addr):
        """Uses binary search to find what module a given address resides in.

        This is much faster than a series of linear checks if you have
        to do it many times. Note that modlist and mod_addrs must be sorted
        in order of the module base address."""
        if self.mod_lookup is None:
            self._make_cache()

        addr = int(addr)
        pos = bisect.bisect_right(self.modlist, addr) - 1
        if pos == -1:
            return obj.NoneObject("Unknown")
        mod = self.mod_lookup[self.modlist[pos]]

        if (addr >= mod.DllBase.v() and
            addr < mod.DllBase.v() + mod.SizeOfImage.v()):
            return mod

        return obj.NoneObject("Unknown")

    def render(self, renderer):
        renderer.table_header([("Offset (V)", "offset_v", "[addrpad]"),
                               ("Name", "file_name", "20"),
                               ('Base', "module_base", "[addrpad]"),
                               ('Size', "module_size", "[addr]"),
                               ('File', "path", "")
                               ])

        for module in self.lsmod():
            renderer.table_row(module.obj_offset,
                               module.BaseDllName,
                               module.DllBase,
                               module.SizeOfImage,
                               module.FullDllName)



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#

__author__ = "Michael Cohen <scudette@gmail.com>"

import re

from rekall import registry
from rekall import constants


class BaseScanner(object):
    """ A more thorough scanner which checks every byte """

    __metaclass__ = registry.MetaclassRegistry

    checks = []
    def __init__(self, profile=None, address_space=None, window_size=8,
                 session=None):
        """The base scanner.

        Args:
           profile: The kernel profile to use for this scan.
           address_space: The address space we use for scanning.
           window_size: The size of the overlap window between each buffer read.
        """
        self.address_space = address_space
        self.window_size = window_size
        self.constraints = None
        self.profile = profile or session.profile
        self.max_length = None
        self.base_offset = None
        self.session = session

    def build_constraints(self):
        self.constraints = []
        for class_name, args in self.checks:
            check = ScannerCheck.classes[class_name](
                profile=self.profile, address_space=self.address_space, **args)
            self.constraints.append(check)

        self.skippers = [c for c in self.constraints if hasattr(c, "skip")]
        self.hits = None

    def check_addr(self, offset):
        """Calls our constraints on the offset and returns if any contraints did
        not match.

        Args:
           offset: The offset to test (in self.address_space).
        """
        for check in self.constraints:
            # Ask the check if this offset is possible.
            val = check.check(offset)
            if not val:
                return False

        return True

    def skip(self, data, data_offset, base_offset=None):
        """Skip uninteresting regions.

        Where should we go next? By default we go 1 byte ahead, but if some of
        the checkers have skippers, we may actually go much farther. Checkers
        with skippers basically tell us that there is no way they can match
        anything before the skipped result, so there is no point in trying them
        on all the data in between. This optimization is useful to really speed
        things up.
        """
        skip = 1
        for s in self.skippers:
            skip_value = s.skip(data, data_offset + skip,
                                base_offset=base_offset)
            skip = max(skip, skip_value)

        return skip

    overlap = 1024
    def scan(self, offset=0, maxlen=None):
        """Scan the region from offset for maxlen.

        Args:
          offset: The starting offset in our current address space to scan.

          maxlen: The maximum length to scan. If no provided we just scan until
            there is no data.

        Yields:
          offsets where all the constrainst are satisfied.
        """
        if maxlen is None:
            last_range = list(self.address_space.get_available_addresses())[-1]
            maxlen = last_range[0] + last_range[1]

        # Delay building the constraints so they can be added after scanner
        # construction.
        if self.constraints is None:
            self.build_constraints()

        # Start scanning from offset until maxlen:
        i = offset

        data_offset = 0
        data = ""

        while i < offset + maxlen:
            # Update the progress bar.
            if self.session:
                self.session.report_progress(
                    "Scanning 0x%08X with %s" % (i, self.__class__.__name__))

            # Check the current offset for a match.
            if self.check_addr(i):
                yield i

            # Allow us to skip uninteresting regions (default skip is 1).
            if data_offset + self.overlap >= len(data):
                # Current region is not valid.
                if not self.address_space.is_valid_address(i):
                    break

                to_read = min(constants.SCAN_BLOCKSIZE, maxlen - (i - offset))

                # Refresh the data buffer.
                data = self.address_space.read(i, to_read)
                data_offset = 0

            # First check if we can skip this point.
            skip = self.skip(data, data_offset, base_offset=i)
            data_offset += skip
            i += skip


class DiscontigScanner(object):
    """A Mixin for Discontiguous scanning."""

    def scan(self, offset=0, maxlen=None):
        maxlen = maxlen or self.profile.get_constant("MaxPointer")

        for (start, length) in self.address_space.get_address_ranges(
            offset, offset + maxlen):
            for match in super(DiscontigScanner, self).scan(start, length):
                yield match


class PointerScanner(DiscontigScanner, BaseScanner):
    """Scan for a bunch of pointers at the same time.

    This scanner takes advantage of the fact that usually the most significant
    bytes of a group of pointers is the same. This common part is scanned for
    first, thereby taking advantage of the scanner skippers.
    """
    def __init__(self, pointers=None, **kwargs):
        """Creates the Pointer Scanner.

        Args:
          pointers: A list of Pointer objects, or simply memory addresses. This
            scanner finds direct references to these addresses in memory.
        """
        super(PointerScanner, self).__init__(**kwargs)

        # The size of a pointer depends on the profile.
        self.address_size = self.profile.get_obj_size("address")
        self.needles = []

        # Find the common string between all the addresses.
        for address in pointers:
            # Encode the address as a pointer according to the current profile.
            tmp = self.profile.address()
            tmp.write(address)

            self.needles.append(tmp.obj_vm.read(0, tmp.size()))

        # The common string between all the needles.
        self.common = self.FindCommonString(self.needles)
        self.checks = [
            ('StringCheck', dict(needle=self.common)),
            ]

    def FindCommonString(self, needles):
        """Find the largest common suffix among all the needles.

        Note we assume all the needles are the same size and pointers are little
        endian (so we work from the end of the string to the beginning).
        """
        common = ""
        for i in range(1, self.address_size):
            possible_match = None
            for needle in needles:
                # If this does not match we stop early.
                if possible_match is not None and possible_match != needle[-i]:
                    return common

                possible_match = needle[-i]

            common = possible_match + common

        return common

    def scan(self, **kwargs):
        for hit in super(PointerScanner, self).scan(**kwargs):
            # Correct the hit for the common suffix.
            hit -= self.address_size - len(self.common)
            data = self.address_space.read(hit, self.address_size)

            if data in self.needles:
                yield hit


class ScannerCheck(object):
    """ A scanner check is a special class which is invoked on an AS to check
    for a specific condition.

    The main method is def check(self, offset):
    This will return True if the condition is true or False otherwise.

    This class is the base class for all checks.
    """

    __metaclass__ = registry.MetaclassRegistry
    __abstract = True

    def __init__(self, profile=None, address_space=None, **_kwargs):
        """The profile that this scanner check should use."""
        self.profile = profile
        self.address_space = address_space

    def object_offset(self, offset):
        return offset

    def check(self, offset):
        _ = offset
        return False

    def skip(self, data, offset, base_offset=None):
        """Determine how many bytes we can skip.

        If you want to speed up the scanning define this method - it
        will be used to skip the data which is obviously not going to
        match. You will need to return the number of bytes from offset
        to skip to. We take the maximum number of bytes to guarantee
        that all checks have a chance of passing.

        Args:
          data: A data buffer we can examine to check for skipping.
          offset: The offset within the data buffer to look at.
          base_offset: The data buffer represents this offset within
            self.address_space.

        Returns:
          number of bytes to be skipped.
        """
        _ = data
        _ = offset
        _ = base_offset
        return 0


class MultiStringFinderCheck(ScannerCheck):
    """A scanner checker for multiple strings."""

    def __init__(self, needles=None, **kwargs):
        """
        Args:
          needles: A list of strings we search for.
        """
        super(MultiStringFinderCheck, self).__init__(**kwargs)
        if not needles:
            needles = []
        self.needles = needles
        self.maxlen = 0
        for needle in needles:
            self.maxlen = max(self.maxlen, len(needle))
        if not self.maxlen:
            raise RuntimeError("No needles of any length were found for the "
                               "MultiStringFinderCheck")

    def check(self, offset):
        verify = self.address_space.read(offset, self.maxlen)

        for match in self.needles:
            if verify[:len(match)] == match:
                return True

        return False

    def skip(self, data, offset, base_offset=None):
        nextval = len(data)
        for needle in self.needles:
            dindex = data.find(needle, offset + 1)
            if dindex > -1:
                nextval = min(nextval, dindex + 1)

        return nextval - offset


class StringCheck(ScannerCheck):
    maxlen = 100

    def __init__(self, needle=None, **kwargs):
        super(StringCheck, self).__init__(**kwargs)
        self.needle = needle

    def check(self, offset):
        return self.address_space.read(offset, len(self.needle)) == self.needle

    def skip(self, data, offset, base_offset=None):
        dindex = data.find(self.needle, offset + 1)
        if dindex > -1:
            return dindex + 1 - offset

        return len(data) - offset


class RegexCheck(ScannerCheck):
    """This check can be quite slow."""
    maxlen = 100

    def __init__(self, regex=None, **kwargs):
        super(RegexCheck, self).__init__(**kwargs)
        self.regex = re.compile(regex)

    def check(self, offset):
        verify = self.address_space.read(offset, self.maxlen)
        return bool(self.regex.match(verify))

    def skip(self, data, offset, base_offset=None):
        m = self.regex.search(data[offset:])
        if m:
            return m.start() + 1

        return len(data) - offset


class ScannerGroup(BaseScanner):
    """Runs a bunch of scanners in one pass over the image."""

    def __init__(self, scanners=None, **kwargs):
        """Create a new scanner group.

        Args:
          scanners: A dict of BaseScanner instances. Keys will be used to refer
          to the scanner, while the value is the scanner instance.
        """
        super(ScannerGroup, self).__init__(**kwargs)
        self.scanners = scanners
        for scanner in scanners.values():
            scanner.address_space = self.address_space

        # A dict to hold all hits for each scanner.
        self.result = {}

    def scan(self, offset=0, maxlen=None):
        available_length = maxlen or self.profile.get_constant("MaxPointer")

        while available_length > 0:
            to_read = min(constants.SCAN_BLOCKSIZE + self.overlap,
                          available_length)

            # Now feed all the scanners from the same address space.
            for name, scanner in self.scanners.items():
                for hit in scanner.scan(offset=offset, maxlen=to_read):
                    # Yield the result as well as cache it.
                    yield name, hit

            # Move to the next scan block.
            offset += constants.SCAN_BLOCKSIZE
            available_length -= constants.SCAN_BLOCKSIZE


class DiscontigScannerGroup(ScannerGroup):
    """A scanner group which works over a virtual address space."""

    def scan(self, offset=0, maxlen=None):
        maxlen = maxlen or self.profile.get_constant("MaxPointer")

        for (start, length) in self.address_space.get_address_ranges(
            offset, offset + maxlen):
            for match in super(DiscontigScannerGroup, self).scan(
                start, maxlen=length):
                yield match


class DebugChecker(ScannerCheck):
    """A check that breaks into the debugger when a condition is met.

    Insert this check inside the check stack and we will break into the debugger
    when all the conditions below us are met.
    """
    def check(self, offset):
        _ = offset
        import pdb; pdb.set_trace() # pylint: disable=multiple-statements
        return True



# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen <scudette@users.sourceforge.net>
# Copyright (c) 2008 Volatile Systems
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

# pylint: disable=protected-access

import copy

from rekall import obj
from rekall import utils

from rekall.plugins.overlays import basic
from rekall.plugins.overlays.windows import pe_vtypes


# Standard vtypes are usually autogenerated by scanning through header
# files, collecting debugging symbol data etc. This file defines
# fixups and improvements to the standard types.
windows_overlay = {
    '_UNICODE_STRING': [None, {
            'Buffer': [None, ['Pointer', dict(
                        target='UnicodeString',
                        target_args=dict(length=lambda x: x.Length)
                        )]],
            }],

    '_EPROCESS' : [None, {
            'CreateTime' : [None, ['WinTimeStamp', {}]],
            'ExitTime' : [None, ['WinTimeStamp', {}]],
            'InheritedFromUniqueProcessId' : [None, ['unsigned int']],
            'ImageFileName' : [None, ['String', dict(length=16)]],
            'UniqueProcessId' : [None, ['unsigned int']],
            }],

    '_ETHREAD' : [None, {
            'CreateTime' : [None, ['ThreadCreateTimeStamp', {}]],
            'ExitTime' : [None, ['WinTimeStamp', {}]],
            }],

    '_OBJECT_SYMBOLIC_LINK' : [None, {
            'CreationTime' : [None, ['WinTimeStamp', {}]],
            }],

    '_KUSER_SHARED_DATA' : [None, {
            'SystemTime' : [None, ['WinTimeStamp', dict(is_utc=True)]],
            'TimeZoneBias' : [None, ['WinTimeStamp', {}]],
            }],

    '_KPCR': [None, {
            # The processor block has varying names between windows versions so
            # we just make them synonyms.
            'ProcessorBlock': lambda x: x.m("Prcb") or x.m("PrcbData"),
            '_IDT': lambda x: x.m("IDT") or x.m("IdtBase"),
            '_GDT': lambda x: x.m("GDT") or x.m("GdtBase"),
            'KdVersionBlock': [None, ['Pointer', dict(
                        target='_KDDEBUGGER_DATA64')]],
            }],

    '_KPRCB': [None, {
            'CurrentThread': [None, ['Pointer', dict(
                        target='_ETHREAD')]],
            'IdleThread': [None, ['Pointer', dict(
                        target='_ETHREAD')]],
            'NextThread': [None, ['Pointer', dict(
                        target='_ETHREAD')]],
            'VendorString': [None, ['String', dict(length=13)]],

            }],

    # The DTB is really an array of 2 ULONG_PTR but we only need the first one
    # which is the value loaded into CR3. The second one, according to procobj.c
    # of the wrk-v1.2, contains the PTE that maps something called hyper space.
    '_KPROCESS' : [None, {
            'DirectoryTableBase' : [None, ['unsigned long']],
    }],

    '_HANDLE_TABLE_ENTRY' : [None, {
            'Object' : [None, ['_EX_FAST_REF']],
            }],

    '_IMAGE_SECTION_HEADER' : [None, {
            'Name' : [0x0, ['String', dict(length=8)]],
            }],

    'PO_MEMORY_IMAGE' : [None, {
            'Signature':   [None, ['String', dict(length=4)]],
            'SystemTime' : [None, ['WinTimeStamp', {}]],
            }],

    '_DBGKD_GET_VERSION64' : [None, {
            'DebuggerDataList' : [None, ['pointer', ['unsigned long']]],
            }],

    '_CM_KEY_NODE' : [None, {
            'Signature' : [None, ['String', dict(length=2)]],
            'LastWriteTime' : [None, ['WinTimeStamp', {}]],
            'Name' : [None, ['String', dict(length=lambda x: x.NameLength)]],
            }],

    '_CM_NAME_CONTROL_BLOCK' : [None, {
            'Name' : [None, ['String', dict(length=lambda x: x.NameLength)]],
            }],

    '_CHILD_LIST' : [None, {
            'List' : [None, ['pointer', ['array', lambda x: x.Count,
                                          ['pointer', ['_CM_KEY_VALUE']]]]],
            }],

    '_CM_KEY_VALUE' : [None, {
            'Signature' : [None, ['String', dict(length=2)]],
            'Name' : [None, ['String', dict(length=lambda x: x.NameLength)]],
            }],

    '_CM_KEY_INDEX' : [None, {
            'Signature' : [None, ['String', dict(length=2)]],
            'List' : [None, ["Array", dict(
                        count=lambda x: x.Count.v() * 2,
                        target="Pointer",
                        target_args=dict(
                            target='_CM_KEY_NODE'
                            )
                        )]],
            }],

    '_TOKEN' : [None, {
            'UserAndGroups' : [None, ['Pointer', dict(
                        target='Array',
                        target_args=dict(
                            count=lambda x: x.UserAndGroupCount,
                            target='_SID_AND_ATTRIBUTES'
                            )
                        )]],
            }],

    '_SID' : [None, {
            'SubAuthority' : [None, ['Array', dict(
                        count=lambda x: x.SubAuthorityCount,
                        target='unsigned long')]],
            }],

    '_CLIENT_ID': [None, {
            'UniqueProcess' : [None, ['unsigned int']],
            'UniqueThread' : [None, ['unsigned int']],
            }],

    '_MMVAD_FLAGS': [None, {
            # Vad Protections. Also known as page protections. The
            # _MMVAD_FLAGS.Protection, 3-bits, is an index into
            # nt!MmProtectToValue (the following list).
            'ProtectionEnum': lambda x: basic.Enumeration(
                choices={
                    0: 'NOACCESS',
                    1: 'READONLY',
                    2: 'EXECUTE',
                    3: 'EXECUTE_READ',
                    4: 'READWRITE',
                    5: 'WRITECOPY',
                    6: 'EXECUTE_READWRITE',
                    7: 'EXECUTE_WRITECOPY',
                    8: 'NOACCESS',
                    9: 'NOCACHE | READONLY',
                    10:'NOCACHE | EXECUTE',
                    11:'NOCACHE | EXECUTE_READ',
                    12:'NOCACHE | READWRITE',
                    13:'NOCACHE | WRITECOPY',
                    14:'NOCACHE | EXECUTE_READWRITE',
                    15:'NOCACHE | EXECUTE_WRITECOPY',
                    16:'NOACCESS',
                    17:'GUARD | READONLY',
                    18:'GUARD | EXECUTE',
                    19:'GUARD | EXECUTE_READ',
                    20:'GUARD | READWRITE',
                    21:'GUARD | WRITECOPY',
                    22:'GUARD | EXECUTE_READWRITE',
                    23:'GUARD | EXECUTE_WRITECOPY',
                    24:'NOACCESS',
                    25:'WRITECOMBINE | READONLY',
                    26:'WRITECOMBINE | EXECUTE',
                    27:'WRITECOMBINE | EXECUTE_READ',
                    28:'WRITECOMBINE | READWRITE',
                    29:'WRITECOMBINE | WRITECOPY',
                    30:'WRITECOMBINE | EXECUTE_READWRITE',
                    31:'WRITECOMBINE | EXECUTE_WRITECOPY',
                    },
                value=x.m("Protection"), name=x.obj_name, type_name=x.obj_type),

            # Vad Types. The _MMVAD_SHORT.u.VadFlags (_MMVAD_FLAGS) struct on XP
            # has individual flags, 1-bit each, for these types. The
            # _MMVAD_FLAGS for all OS after XP has a member of the
            # _MMVAD_FLAGS.VadType, 3-bits, which is an index into the following
            # enumeration.
            "VadTypeEnum": lambda x: basic.Enumeration(
                choices={
                    0: 'VadNone',
                    1: 'VadDevicePhysicalMemory',
                    2: 'VadImageMap',
                    3: 'VadAwe',
                    4: 'VadWriteWatch',
                    5: 'VadLargePages',
                    6: 'VadRotatePhysical',
                    7: 'VadLargePageSection',
                    },
                value=x.m("VadType"), name=x.obj_name, type_name=x.obj_type),
            }],

    # The environment is a null termionated _UNICODE_STRING array. Print with
    # list(eprocess.Peb.ProcessParameters.Environment)
    '_RTL_USER_PROCESS_PARAMETERS': [None, {
            'Environment': [None, ['Pointer', {
                        'target': 'ListArray',
                        'target_args': {
                            'target': "UnicodeString"
                            }
                        }]],
            }],

    '_DEVICE_OBJECT': [None, {
            'DeviceType': [None, ['Enumeration', dict(choices={
                            0x00000027 : 'FILE_DEVICE_8042_PORT',
                            0x00000032 : 'FILE_DEVICE_ACPI',
                            0x00000029 : 'FILE_DEVICE_BATTERY',
                            0x00000001 : 'FILE_DEVICE_BEEP',
                            0x0000002a : 'FILE_DEVICE_BUS_EXTENDER',
                            0x00000002 : 'FILE_DEVICE_CD_ROM',
                            0x00000003 : 'FILE_DEVICE_CD_ROM_FILE_SYSTEM',
                            0x00000030 : 'FILE_DEVICE_CHANGER',
                            0x00000004 : 'FILE_DEVICE_CONTROLLER',
                            0x00000005 : 'FILE_DEVICE_DATALINK',
                            0x00000006 : 'FILE_DEVICE_DFS',
                            0x00000035 : 'FILE_DEVICE_DFS_FILE_SYSTEM',
                            0x00000036 : 'FILE_DEVICE_DFS_VOLUME',
                            0x00000007 : 'FILE_DEVICE_DISK',
                            0x00000008 : 'FILE_DEVICE_DISK_FILE_SYSTEM',
                            0x00000033 : 'FILE_DEVICE_DVD',
                            0x00000009 : 'FILE_DEVICE_FILE_SYSTEM',
                            0x0000003a : 'FILE_DEVICE_FIPS',
                            0x00000034 : 'FILE_DEVICE_FULLSCREEN_VIDEO',
                            0x0000000a : 'FILE_DEVICE_INPORT_PORT',
                            0x0000000b : 'FILE_DEVICE_KEYBOARD',
                            0x0000002f : 'FILE_DEVICE_KS',
                            0x00000039 : 'FILE_DEVICE_KSEC',
                            0x0000000c : 'FILE_DEVICE_MAILSLOT',
                            0x0000002d : 'FILE_DEVICE_MASS_STORAGE',
                            0x0000000d : 'FILE_DEVICE_MIDI_IN',
                            0x0000000e : 'FILE_DEVICE_MIDI_OUT',
                            0x0000002b : 'FILE_DEVICE_MODEM',
                            0x0000000f : 'FILE_DEVICE_MOUSE',
                            0x00000010 : 'FILE_DEVICE_MULTI_UNC_PROVIDER',
                            0x00000011 : 'FILE_DEVICE_NAMED_PIPE',
                            0x00000012 : 'FILE_DEVICE_NETWORK',
                            0x00000013 : 'FILE_DEVICE_NETWORK_BROWSER',
                            0x00000014 : 'FILE_DEVICE_NETWORK_FILE_SYSTEM',
                            0x00000028 : 'FILE_DEVICE_NETWORK_REDIRECTOR',
                            0x00000015 : 'FILE_DEVICE_NULL',
                            0x00000016 : 'FILE_DEVICE_PARALLEL_PORT',
                            0x00000017 : 'FILE_DEVICE_PHYSICAL_NETCARD',
                            0x00000018 : 'FILE_DEVICE_PRINTER',
                            0x00000019 : 'FILE_DEVICE_SCANNER',
                            0x0000001c : 'FILE_DEVICE_SCREEN',
                            0x00000037 : 'FILE_DEVICE_SERENUM',
                            0x0000001a : 'FILE_DEVICE_SERIAL_MOUSE_PORT',
                            0x0000001b : 'FILE_DEVICE_SERIAL_PORT',
                            0x00000031 : 'FILE_DEVICE_SMARTCARD',
                            0x0000002e : 'FILE_DEVICE_SMB',
                            0x0000001d : 'FILE_DEVICE_SOUND',
                            0x0000001e : 'FILE_DEVICE_STREAMS',
                            0x0000001f : 'FILE_DEVICE_TAPE',
                            0x00000020 : 'FILE_DEVICE_TAPE_FILE_SYSTEM',
                            0x00000038 : 'FILE_DEVICE_TERMSRV',
                            0x00000021 : 'FILE_DEVICE_TRANSPORT',
                            0x00000022 : 'FILE_DEVICE_UNKNOWN',
                            0x0000002c : 'FILE_DEVICE_VDM',
                            0x00000023 : 'FILE_DEVICE_VIDEO',
                            0x00000024 : 'FILE_DEVICE_VIRTUAL_DISK',
                            0x00000025 : 'FILE_DEVICE_WAVE_IN',
                            0x00000026 : 'FILE_DEVICE_WAVE_OUT',
                            })]],
            }],
    '_DRIVER_OBJECT': [None, {
            'MajorFunction': [None, ['IndexedArray', dict(
                        index_table={
                            'IRP_MJ_CREATE': 0,
                            'IRP_MJ_CREATE_NAMED_PIPE': 1,
                            'IRP_MJ_CLOSE': 2,
                            'IRP_MJ_READ': 3,
                            'IRP_MJ_WRITE': 4,
                            'IRP_MJ_QUERY_INFORMATION': 5,
                            'IRP_MJ_SET_INFORMATION': 6,
                            'IRP_MJ_QUERY_EA': 7,
                            'IRP_MJ_SET_EA': 8,
                            'IRP_MJ_FLUSH_BUFFERS': 9,
                            'IRP_MJ_QUERY_VOLUME_INFORMATION': 10,
                            'IRP_MJ_SET_VOLUME_INFORMATION': 11,
                            'IRP_MJ_DIRECTORY_CONTROL': 12,
                            'IRP_MJ_FILE_SYSTEM_CONTROL': 13,
                            'IRP_MJ_DEVICE_CONTROL': 14,
                            'IRP_MJ_INTERNAL_DEVICE_CONTROL': 15,
                            'IRP_MJ_SHUTDOWN': 16,
                            'IRP_MJ_LOCK_CONTROL': 17,
                            'IRP_MJ_CLEANUP': 18,
                            'IRP_MJ_CREATE_MAILSLOT': 19,
                            'IRP_MJ_QUERY_SECURITY': 20,
                            'IRP_MJ_SET_SECURITY': 21,
                            'IRP_MJ_POWER': 22,
                            'IRP_MJ_SYSTEM_CONTROL': 23,
                            'IRP_MJ_DEVICE_CHANGE': 24,
                            'IRP_MJ_QUERY_QUOTA': 25,
                            'IRP_MJ_SET_QUOTA': 26,
                            'IRP_MJ_PNP': 27
                            },
                        target="Pointer",
                        target_args=dict(target="Function"),
                        )]],
            }],

    # This defines _PSP_CID_TABLE as an alias for _HANDLE_TABLE.
    "_PSP_CID_TABLE": "_HANDLE_TABLE"
}


class _UNICODE_STRING(obj.Struct):
    """Class representing a _UNICODE_STRING

    Adds the following behavior:
      * The Buffer attribute is presented as a Python string rather
        than a pointer to an unsigned short.
      * The __str__ method returns the value of the Buffer.
    """

    def v(self, vm=None):
        length = self.Length.v(vm=vm)
        if length > 0 and length <= 1024:
            data = self.Buffer.dereference_as(
                'UnicodeString',
                target_args=dict(
                    length=length),
                vm=vm)
            return data.v()
        else:
            return ''

    def __nonzero__(self):
        ## Unicode strings are valid if they point at a valid memory
        return bool(self.Buffer)

    def __eq__(self, other):
        return unicode(self) == utils.SmartUnicode(other)

    def __str__(self):
        return self.v() or ""

    def write(self, string):
        self.Buffer.dereference().write(string)
        self.Length = len(string) * 2


class _EPROCESS(obj.Struct):
    """ An extensive _EPROCESS with bells and whistles """

    @property
    def Peb(self):
        """ Returns a _PEB object which is using the process address space.

        The PEB structure is referencing back into the process address
        space so we need to switch address spaces when we look at
        it. This method ensure this happens automatically.
        """
        return self.m("Peb").dereference_as(
            vm=self.get_process_address_space())

    @property
    def IsWow64(self):
        """Returns True if this is a wow64 process"""
        return hasattr(self, 'Wow64Process') and self.Wow64Process.v() != 0

    @property
    def SessionId(self):
        """Returns the Session ID of the process"""

        if self.Session.is_valid():
            process_space = self.get_process_address_space()
            if process_space:
                return self.obj_profile._MM_SESSION_SPACE(
                    offset=self.Session, vm=process_space).SessionId

        return obj.NoneObject("Cannot find process session")

    def get_process_address_space(self):
        """ Gets a process address space for a task given in _EPROCESS """
        directory_table_base = self.Pcb.DirectoryTableBase.v()

        try:
            process_as = self.obj_vm.__class__(
                base=self.obj_vm.base, session=self.obj_vm.session,
                dtb=directory_table_base)
        except AssertionError, e:
            return obj.NoneObject("Unable to get process AS: %s" % e)

        process_as.name = "Process {0}".format(self.UniqueProcessId)

        return process_as

    def _get_modules(self, the_list, the_type):
        """Generator for DLLs in one of the 3 PEB lists"""
        if self.UniqueProcessId and the_list:
            for l in the_list.list_of_type("_LDR_DATA_TABLE_ENTRY", the_type):
                yield l

    def get_init_modules(self):
        return self._get_modules(self.Peb.Ldr.InInitializationOrderModuleList,
                                 "InInitializationOrderLinks")

    def get_mem_modules(self):
        return self._get_modules(self.Peb.Ldr.InMemoryOrderModuleList,
                                 "InMemoryOrderLinks")

    def get_load_modules(self):
        return self._get_modules(
            self.Peb.Ldr.InLoadOrderModuleList, "InLoadOrderLinks")

    def get_token(self):
        """Return the process's TOKEN object if its valid"""

        # The dereference checks if the address is valid
        # and returns obj.NoneObject if it fails
        token = self.Token.dereference_as("_TOKEN")

        # This check fails if the above dereference failed
        # or if any of the _TOKEN specific validity tests failed.
        if token.is_valid():
            return token

        return obj.NoneObject("Cannot get process Token")

    def ObReferenceObjectByHandle(self, handle, type=None):
        """Search the object table and retrieve the object by handle.

        Args:
          handle: The handle we search for.
          type: The object will be cast to this type.
        """
        for h in self.ObjectTable.handles():
            if h.HandleValue == handle:
                if type is None:
                    return h
                else:
                    return h.dereference_as(type)

        return obj.NoneObject("Could not find handle in ObjectTable")



class _POOL_HEADER(obj.Struct):
    """Extension to support retrieving allocations inside the pool."""

    def get_rounded_size(self, object_name):
        """Returns the size of the object accounting for pool alignment."""
        size_of_obj = self.obj_profile.get_obj_size(object_name)
        pool_align = self.obj_profile.get_constant("PoolAlignment")

        # Size is rounded to pool alignment
        extra = size_of_obj % pool_align
        if extra:
            size_of_obj += pool_align - extra

        return size_of_obj

    def end(self):
        return self.obj_offset + self.size()

    def get_object(self, object_name, allocations):
        """This implements retrieving an object from the pool allocation using
        the "Bottom Up" method.  NOTE: This method does not work on windows 8
        since allocations are rounded up to a fixed size. Newer versions of
        windows have different _POOL_HEADER implementations.

        The following is provided by MHL:

        For example, let's assume the following object has no preamble, then
        we'd take the base of pool header and add the size of pool header to
        reach the base of the object. Layout in memory looks like this:


        _POOL_HEADER
        <TheObject>

        Now let's assume the object has a preamble - an _OBJECT_HEADER with no
        optional headers.

        _POOL_HEADER
        _OBJECT_HEADER
        <TheObject>

        Its easy to calculate the offset of the object, because you always know
        the size of _POOL_HEADER and _OBJECT_HEADER. However, one situation
        complicates this calculation. There may be optional headers between the
        pool header and object header like this:

        _POOL_HEADER
        <SomeHeaderA>
        <SomeHeaderB>
        _OBJECT_HEADER
        <TheObject>

        The _OBJECT_HEADER itself is the "map" which tell us how many optional
        headers there are. The question becomes - how do we find the
        _OBJECT_HEADER when the very information we need (distance between pool
        header and object header) is stored in the _OBJECT_HEADER? Furthermore,
        we can't statically set preambles, because not only do they differ
        between objects (i.e. mutants may have different optional headers than
        file objects), but they sometimes differ between objects of the same
        type (for example one process may have 2 optional headers and another
        process may only have 1). That flexibility is not really possible with
        the preambles - at least how they were implemented at the time of these
        changes.

        So the "bottom up" approach takes into account two values which *are*
        reliable:

        1. The size of the pool (_POOL_HEADER.BlockSize)
        2. The size of the object you expect to find in the pool
           (i.e. get_obj_size("_EPROCESS"))

        So with that information, you can find the end of the pool
        (i.e. starting from the bottom), subtract the size of the object
        (working our way up), and then you've got the offset of the
        object. Always, the _OBJECT_HEADER (if there is one) directly precedes
        the object, so once you've got the object's offset, you can find the
        _OBJECT_HEADER. And from there, since _OBJECT_HEADER is the "map" you
        can find any optional headers.

        Args:
          name: The name of the object type to retrieve. Note: name must be
            allocations.

          allocations: The list of objects which form this allocation.
        """
        pool_align = self.obj_profile.get_constant("PoolAlignment")

        # We start at the end of the allocation, and go backwards for each
        # object.
        offset = self.obj_offset + self.BlockSize * pool_align

        for name in reversed(allocations):
            # Rewind to the start of this object.
            offset -= self.get_rounded_size(name)

            # Make a new object instance.
            obj = self.obj_profile.Object(name, vm=self.obj_vm, offset=offset)
            if name == object_name:
                return obj

            # Rewind past the object's preamble
            offset -= obj.preamble_size()

        raise KeyError("object not present in preamble.")

    @property
    def FreePool(self):
        return self.PoolType.v() == 0

    @property
    def NonPagedPool(self):
        return self.PoolType.v() % 2 == 1

    @property
    def PagedPool(self):
        return self.PoolType.v() % 2 == 0 and self.PoolType.v() > 0


class _TOKEN(obj.Struct):
    """A class for Tokens"""

    def is_valid(self):
        """Override BaseObject.is_valid with some additional
        checks specific to _TOKEN objects."""
        return (super(_TOKEN, self).is_valid() and
                self.TokenInUse in (0, 1) and self.SessionId < 10)

    def get_sids(self):
        """Generator for process SID strings"""
        if self.UserAndGroupCount < 0xFFFF:
            for sa in self.UserAndGroups.dereference():
                sid = sa.Sid.dereference_as('_SID')
                for i in sid.IdentifierAuthority.Value:
                    id_auth = i
                yield "S-" + "-".join(str(i) for i in (sid.Revision, id_auth) +
                                      tuple(sid.SubAuthority))


class _ETHREAD(obj.Struct):
    """ A class for threads """

    def owning_process(self):
        """Return the EPROCESS that owns this thread"""
        return self.Tcb.ApcState.Process.dereference_as("_EPROCESS")

    def attached_process(self):
        """Return the EPROCESS that this thread is currently
        attached to."""
        return self.Tcb.ApcState.Process.dereference_as("_EPROCESS")


class _HANDLE_TABLE(obj.Struct):
    """ A class for _HANDLE_TABLE.

    This used to be a member of _EPROCESS but it was isolated per issue
    91 so that it could be subclassed and used to service other handle
    tables, such as the _KDDEBUGGER_DATA64.PspCidTable.
    """

    def get_item(self, entry, handle_value = 0):
        """Returns the OBJECT_HEADER of the associated handle. The parent
        is the _HANDLE_TABLE_ENTRY so that an object can be linked to its
        GrantedAccess.
        """
        return entry.Object.dereference_as("_OBJECT_HEADER", parent = entry,
                                           handle_value = handle_value)

    def _make_handle_array(self, offset, level, depth = 0):
        """ Returns an array of _HANDLE_TABLE_ENTRY rooted at offset,
        and iterates over them.
        """
        # The counts below are calculated by taking the size of a page and dividing
        # by the size of the data type contained within the page. For more information
        # see http://blogs.technet.com/b/markrussinovich/archive/2009/09/29/3283844.aspx
        if level > 0:
            count = 0x1000 / self.obj_profile.get_obj_size("address")
            target = "address"
        else:
            count = 0x1000 / self.obj_profile.get_obj_size("_HANDLE_TABLE_ENTRY")
            target = "_HANDLE_TABLE_ENTRY"

        table = self.obj_profile.Array(
            offset = offset, vm = self.obj_vm,
            count = count, target = target, parent = self)

        if table:
            for entry in table:
                if not entry.is_valid():
                    break

                if level > 0:
                    ## We need to go deeper:
                    for h in self._make_handle_array(entry, level - 1, depth):
                        yield h
                    depth += 1
                else:

                    # All handle values are multiples of four, on both x86 and x64.
                    handle_multiplier = 4
                    # Calculate the starting handle value for this level.
                    handle_level_base = depth * count * handle_multiplier
                    # The size of a handle table entry.
                    handle_entry_size = self.obj_profile.get_obj_size("_HANDLE_TABLE_ENTRY")
                    # Finally, compute the handle value for this object.
                    handle_value = ((entry.obj_offset - table[0].obj_offset) /
                                   (handle_entry_size / handle_multiplier)) + handle_level_base

                    ## OK We got to the bottom table, we just resolve
                    ## objects here:
                    item = self.get_item(entry, handle_value)

                    if item == None:
                        continue

                    try:
                        # New object header
                        if item.TypeIndex != 0x0:
                            yield item
                    except AttributeError:
                        if item.Type.Name:
                            yield item

    def handles(self):
        """ A generator which yields this process's handles

        _HANDLE_TABLE tables are multi-level tables at the first level
        they are pointers to second level table, which might be
        pointers to third level tables etc, until the final table
        contains the real _OBJECT_HEADER table.

        This generator iterates over all the handles recursively
        yielding all handles. We take care of recursing into the
        nested tables automatically.
        """
        # This should work equally for 32 and 64 bit systems
        LEVEL_MASK = 7

        TableCode = self.TableCode.v() & ~LEVEL_MASK
        table_levels = self.TableCode.v() & LEVEL_MASK
        offset = TableCode

        for h in self._make_handle_array(offset, table_levels):
            yield h


class _PSP_CID_TABLE(_HANDLE_TABLE):
    """Subclass the Windows handle table object for parsing PspCidTable"""

    def get_item(self, entry, handle_value = 0):
        p = self.obj_profile.Object("address", entry.Object.v(), self.obj_vm)

        handle = self.obj_profile.Object(
            "_OBJECT_HEADER",
            offset=(p & ~7) - self.obj_profile.get_obj_offset(
                '_OBJECT_HEADER', 'Body'),
            vm = self.obj_vm)

        return handle


class _OBJECT_HEADER(obj.Struct):
    """A Rekall Memory Forensics object to handle Windows object headers.

    This object applies only to versions below windows 7.
    """

    optional_headers = [
        ('NameInfo', 'NameInfoOffset', '_OBJECT_HEADER_NAME_INFO'),
        ('HandleInfo', 'HandleInfoOffset', '_OBJECT_HEADER_HANDLE_INFO'),
        ('HandleInfo', 'QuotaInfoOffset', '_OBJECT_HEADER_QUOTA_INFO')]

    def __init__(self, handle_value=0, **kwargs):
        self.HandleValue = handle_value
        self._preamble_size = 0
        super(_OBJECT_HEADER, self).__init__(**kwargs)

        # Create accessors for optional headers
        self.find_optional_headers()

    def find_optional_headers(self):
        """Find this object's optional headers."""
        offset = self.obj_offset

        for name, name_offset, objtype in self.optional_headers:
            if self.obj_profile.has_type(objtype):
                header_offset = self.m(name_offset).v()
                if header_offset:
                    o = self.obj_profile.Object(type_name=objtype,
                                                offset=offset - header_offset,
                                                vm = self.obj_vm)
                else:
                    o = obj.NoneObject("Header not set")

                setattr(self, name, o)

                # Optional headers stack before this _OBJECT_HEADER.
                if o:
                    self._preamble_size += o.size()

    def preamble_size(self):
        return self._preamble_size

    def size(self):
        """The size of the object header is actually the position of the Body
        element."""
        return self.obj_profile.get_obj_offset("_OBJECT_HEADER", "Body")

    @property
    def GrantedAccess(self):
        if self.obj_parent:
            return self.obj_parent.GrantedAccess
        return obj.NoneObject("No parent known")

    def dereference_as(self, type_name, vm=None):
        """Instantiate an object from the _OBJECT_HEADER.Body"""
        return self.obj_profile.Object(type_name=type_name, offset=self.Body.obj_offset,
                                       vm=vm or self.obj_vm, parent=self)

    def get_object_type(self, vm=None):
        """Return the object's type as a string"""
        type_obj = self.obj_profile._OBJECT_TYPE(vm=vm or self.obj_vm,
                                                 offset=self.Type)

        return type_obj.Name.v()



class _FILE_OBJECT(obj.Struct):
    """Class for file objects"""

    @property
    def AccessString(self):
        """Make a nicely formatted ACL string."""
        return (((self.ReadAccess > 0 and "R") or '-') +
                ((self.WriteAccess > 0  and "W") or '-') +
                ((self.DeleteAccess > 0 and "D") or '-') +
                ((self.SharedRead > 0 and "r") or '-') +
                ((self.SharedWrite > 0 and "w") or '-') +
                ((self.SharedDelete > 0 and "d") or '-'))

    def file_name_with_device(self):
        """Return the name of the file, prefixed with the name
        of the device object to which the file belongs"""
        name = ""
        if self.DeviceObject:
            object_hdr = self.obj_profile._OBJECT_HEADER(
                offset=(self.DeviceObject.v() - self.obj_profile.get_obj_offset(
                        "_OBJECT_HEADER", "Body")),
                vm=self.obj_vm)

            if object_hdr.NameInfo:
                name = u"\\Device\\{0}".format(object_hdr.NameInfo.Name)

        if self.FileName:
            name += self.FileName.v()

        return name


class _EX_FAST_REF(obj.Struct):
    """This type allows instantiating an object from its .Object member."""

    def __init__(self, target=None, **kwargs):
        self.target = target
        super(_EX_FAST_REF, self).__init__(**kwargs)

    def dereference(self, vm=None):
        if self.target is None:
            raise TypeError("No target specified for dereferencing an _EX_FAST_REF.")

        return self.dereference_as(self.target)

    def dereference_as(self, type_name, parent=None, vm=None, **kwargs):
        """Use the _EX_FAST_REF.Object pointer to resolve an object of the
        specified type.
        """
        MAX_FAST_REF = self.obj_profile.constants['MAX_FAST_REF']
        return self.obj_profile.Object(
            type_name=type_name, offset=self.m("Object").v() & ~MAX_FAST_REF,
            vm=vm or self.obj_vm, parent=parent or self, **kwargs)

    def __getattr__(self, attr):
        return getattr(self.dereference(), attr)


class _CM_KEY_BODY(obj.Struct):
    """Registry key"""

    def full_key_name(self):
        output = []
        kcb = self.KeyControlBlock
        while kcb.ParentKcb:
            if kcb.NameBlock.Name == None:
                break
            output.append(str(kcb.NameBlock.Name))
            kcb = kcb.ParentKcb
        return "\\".join(reversed(output))

class _MMVAD_FLAGS(obj.Struct):
    """This is for _MMVAD_SHORT.u.VadFlags"""

    def __str__(self):
        result = []
        for name in sorted(self.members):
            if name.endswith("Enum"):
                continue

            try:
                attribute = getattr(self, name)
                if attribute.v():
                    result.append("%s: %s" % (name, attribute))
            except AttributeError:
                pass

        return ", ".join(result)

class _MMVAD_FLAGS2(_MMVAD_FLAGS):
    """This is for _MMVAD_LONG.u2.VadFlags2"""
    pass

class _MMSECTION_FLAGS(_MMVAD_FLAGS):
    """This is for _CONTROL_AREA.u.Flags"""
    pass


class VadTraverser(obj.Struct):
    """The windows Vad tree is basically the same in all versions of windows,
    but the exact name of the stucts vary with version. This is the base class
    for all Vad traversor.
    """
    ## The actual type depends on this tag value.
    tag_map = {'Vadl': '_MMVAD_LONG',
               'VadS': '_MMVAD_SHORT',
               'Vad ': '_MMVAD',
               'VadF': '_MMVAD_SHORT',
               'Vadm': '_MMVAD_LONG',
              }

    def traverse(self, visited=None, depth=0):
        """ Traverse the VAD tree by generating all the left items,
        then the right items.

        We try to be tolerant of cycles by storing all offsets visited.
        """
        if depth > 100:
            raise RuntimeError("Vad tree too deep - something went wrong!")

        if visited == None:
            visited = set()

        ## We try to prevent loops here
        if self.obj_offset in visited:
            return

        self.obj_context['depth'] = depth

        # Find out which Vad type we need to be:
        if self.Tag in self.tag_map:
            yield self.cast(self.tag_map[self.Tag])

        # This tag is valid for the Root.
        elif depth and self.Tag.v() != "\x00":
            return

        for c in self.LeftChild.traverse(visited=visited, depth=depth+1):
            visited.add(self.obj_offset)
            yield c

        for c in self.RightChild.traverse(visited=visited, depth=depth+1):
            visited.add(self.obj_offset)
            yield c


import crash_vtypes
import kdbg_vtypes
import ssdt_vtypes

# Reference:
# http://computer.forensikblog.de/en/2006/03/dmp-file-structure.html

crash_overlays = {
    "_DMP_HEADER": [None, {
            'Signature': [None, ['String', dict(length=4)]],
            'ValidDump': [None, ['String', dict(length=4)]],
            'SystemTime': [None, ['WinTimeStamp']],
            'DumpType': [None, ['Enumeration', {
                        'choices': {
                            1: "Full Dump",
                            2: "Kernel Dump",
                            },
                        'target': 'unsigned int'}]],
            }],
    '_PHYSICAL_MEMORY_DESCRIPTOR' : [ None, {
            'Run' : [ None, ['Array', dict(
                        count=lambda x: x.NumberOfRuns,
                        target='_PHYSICAL_MEMORY_RUN')]],
            }],
    }

crash_overlays['_DMP_HEADER64'] = copy.deepcopy(crash_overlays['_DMP_HEADER'])


class CrashDump32Profile(basic.Profile32Bits, basic.BasicWindowsClasses):
    """A profile for crash dumps."""
    def __init__(self, **kwargs):
        super(CrashDump32Profile, self).__init__(**kwargs)
        self.add_overlay(crash_vtypes.crash_vtypes)
        self.add_overlay(crash_overlays)


class CrashDump64Profile(basic.ProfileLLP64, basic.BasicWindowsClasses):
    """A profile for crash dumps."""
    def __init__(self, **kwargs):
        super(CrashDump64Profile, self).__init__(**kwargs)
        self.add_overlay(crash_vtypes.crash_vtypes)
        self.add_overlay(crash_vtypes.crash_64_vtypes)
        self.add_overlay(crash_overlays)


class KDDebuggerProfile(basic.ProfileLLP64, basic.BasicWindowsClasses):
    """A domain specific profile for the kernel debugger structures.

    The kernel debugger strucutures do not vary with windows operating system
    version very much. This is probably done to make it easier for Windbg to
    support all the different windows versions.
    """

    def __init__(self, **kwargs):
        super(KDDebuggerProfile, self).__init__(**kwargs)

        self.add_types(kdbg_vtypes.kdbg_vtypes)
        self.add_overlay(kdbg_vtypes.kdbg_overlay)
        self.add_classes({
                "_KDDEBUGGER_DATA64": kdbg_vtypes._KDDEBUGGER_DATA64
                })


class BaseWindowsProfile(basic.BasicWindowsClasses):
    """Common symbols for all of windows kernel profiles."""
    _md_os = "windows"

    __abstract = True

    def __init__(self, **kwargs):
        super(BaseWindowsProfile, self).__init__(**kwargs)

        # KDBG types.
        self.add_types(kdbg_vtypes.kdbg_vtypes)
        self.add_overlay(kdbg_vtypes.kdbg_overlay)
        self.add_classes({
                "_KDDEBUGGER_DATA64": kdbg_vtypes._KDDEBUGGER_DATA64
                })

        self.add_types(ssdt_vtypes.ssdt_vtypes)
        self.add_classes({
            '_UNICODE_STRING': _UNICODE_STRING,
            '_EPROCESS': _EPROCESS,
            '_ETHREAD': _ETHREAD,
            '_HANDLE_TABLE': _HANDLE_TABLE,
            '_POOL_HEADER': _POOL_HEADER,
            '_OBJECT_HEADER': _OBJECT_HEADER,
            '_PSP_CID_TABLE': _PSP_CID_TABLE,
            '_FILE_OBJECT': _FILE_OBJECT,
            '_EX_FAST_REF': _EX_FAST_REF,
            '_CM_KEY_BODY': _CM_KEY_BODY,
            '_MMVAD_FLAGS': _MMVAD_FLAGS,
            '_MMVAD_FLAGS2': _MMVAD_FLAGS2,
            '_MMSECTION_FLAGS': _MMSECTION_FLAGS,
            })

        self.add_overlay(windows_overlay)

        # Also apply basic PE file parsing to the overlays.
        pe_vtypes.PEFileImplementation.Modify(self)

        # Pooltags for common objects.
        self.add_constants(DRIVER_POOLTAG="Dri\xf6",
                           EPROCESS_POOLTAG="Pro\xe3",
                           FILE_POOLTAG="Fil\xe5",
                           SYMLINK_POOLTAG="Sym\xe2",
                           MODULE_POOLTAG="MmLd",
                           MUTANT_POOLTAG="Mut\xe1",
                           THREAD_POOLTAG='\x54\x68\x72\xe5',
                           )
